#include "main.h"

#define USER_RL	66

struct user *userTable;
int userTableLast;

void convLookup(unsigned char b3[3], unsigned char b4[4], int n) {
	
	const char lookup[65] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

		
	b4[0] = b3[0] >> 2;
	b4[1] = b3[0] & 0x3;
	b4[1] <<= 4;

	if(n > 1) {
		b4[1] |= (b3[1] >> 4);
		b4[2] = b3[1] & 0xF;
		b4[2] <<= 2;
	}

	if(n > 2) {
		b4[2] |= b3[2]	>> 6;
		b4[3] = b3[2] & 0x3F;
	}
	
	b4[0] = lookup[b4[0]];
	b4[1] = lookup[b4[1]];
	
	if(n == 1) {
		b4[2] = '=';
		b4[3] = '=';
	} 
	else if(n == 2) {
		b4[2] = lookup[b4[2]];
		b4[3] = '=';
	}
	else {
		b4[2] = lookup[b4[2]];
		b4[3] = lookup[b4[3]];
	}
}


char * cryptBase64(const char *userPass, char *cUserPass) {
	char base64[220];
	char b3[3], b4[4];
	char *q;
	int  i, p;

	i = strlen(userPass);
	p = 0;
	q = base64;

	while(i > 3) {
		b3[0] = userPass[p++];
		b3[1] = userPass[p++];
		b3[2] = userPass[p++];
	
		convLookup(b3, b4, 3);
		memcpy(q, b4, 4);
	
		q += 4;
		i -= 3;
	}

	b3[0] = userPass[p++];	
	if(i > 1) 
		b3[1] = userPass[p++];
	if(i > 2)
		b3[2] = userPass[p];

	convLookup(b3, b4, i);
	memcpy(q, b4, 4);
	q[4] = '\0';

	strcpy(cUserPass, crypt(base64, CRYPT_SALT));

	return cUserPass;
}


int storeUserPass(FILE *fp, const char user[33], const char *pass) {
	
	char line[69], cUser[33], userPass[162], cUserPass[35];	
	int n;

	//memcpy(cUser, crypt(user, CRYPT_SALT), 34);
	memcpy(cUser, user, 32);
	
	des_setparity(key);	
	ecb_crypt(key, cUser, 32, DES_ENCRYPT);
	
	if(fseek(fp, INI_USER_OFF, SEEK_SET) == -1) {
		Error("Invalid INI file");
		return 0;
	}

	while(!feof(fp)) {
		n = fread(line, 1, USER_RL, fp);
		if(n == 0) break;
	    if(n != USER_RL)	{
			Error("Invalid INI file");
			return 0;
		}

		if(strncmp(cUser, line, 32) == 0) {
			Error("User name already exists");
			return 0;
		}
	}
	
	if(fwrite(cUser, 1, 32, fp) != 32) {
		Error("Unable to write to INI file");
		return 0;
	}

	strcpy(userPass, user);
	strcat(userPass, ":");
	strcat(userPass, pass);
	
	if(fwrite(cryptBase64(userPass, cUserPass), 1, 34, fp) != 34){
		Error("Unable to write to INI file");
		return 0;
	}
	
	return 1;
}
	

/* add users to the user database */
int adduser(const char *ini_file) {
	FILE *fp;
	char user[33];
	char *p, *password;
	int another = 1;

	if((fp = fopen(ini_file, "r+")) == NULL) {
		perror("fopen");
		return 0;
	}

	while(another) {
		bzero(user, 33);

		fflush(stdin);
		printf("\nUser Name:");
		scanf("%32s", user);
		
		if(user[0] == '\0') {
			printf("\nUser name cannot be empty");
			continue;
		}

		printf("Password:");
		p = getpass("");
		getchar();

		password = (char *)malloc(strlen(p)+1);
		strcpy(password, p);
		storeUserPass(fp, user, password);
		free(password);

		printf("\nAdd another user(Y/N):");
		user[0] = getchar();
//		getchar();
		if(user[0] != 'y' && user[0] != 'Y')
			another = 0;
	}
	fclose(fp);
}

int delRecord(FILE *fp) {
	FILE *ftmp;
	int c;
	unsigned long n, i;

	if((ftmp = tmpfile()) == NULL) 
		return 0;
	
	n = ftell(fp);
	while((c = getc(fp)) != EOF)
		putc(c, ftmp);

	if(ftruncate(fileno(fp), n - USER_RL) == -1)
		return 0;
	
	fseek(fp, 0L, SEEK_END);
	
	fseek(ftmp, 0L, SEEK_SET);	
	while((c = getc(ftmp)) != EOF)
		putc(c, fp);

	fclose(ftmp);

	return 1;
}
	

int removeUser(FILE *fp, const char *user) {
		
	char cUser[33];
	char line[67];
	unsigned n;
	
	if(fseek(fp, INI_USER_OFF, SEEK_SET) == -1) {
		Error("Invalid INI file");
		return 0;
	}
	
	//memcpy(cUser, crypt(user, CRYPT_SALT), 34);
	
	memcpy(cUser, user, 32);
	
	des_setparity(key);	
	ecb_crypt(key, cUser, 32, DES_ENCRYPT);
	
	while(!feof(fp)) {
		n = fread(line, 1, USER_RL, fp);
		if(n == 0) break;
	    if(n != USER_RL)	{
			Error("Invalid INI file");
			return 0;
		}

		if(strncmp(cUser, line, 32) == 0) {
			delRecord(fp);
			return 1;
		}
	}
	
	Error("User not found");
	return 0;
}		
	
int deleteUser(const char *ini_file) {
	FILE *fp;
	char user[33];	
	int another = 1;
	

	if((fp = fopen(ini_file, "r+")) == NULL) {
		perror("fopen");
		return 0;
	}
	
	while(another) {
		bzero(user, 33);
		fflush(stdin);
		printf("\nUser Name:");
		scanf("%32s", user);
		getchar();	
		
		if(user[0] == '\0') {
			printf("\nUser name cannot be empty");
			continue;
		}
		
		removeUser(fp, user);
	
		printf("\nDelete another user(Y/N):");		
		user[0] = getchar();		
		if(user[0] != 'y' && user[0] != 'Y')
			another = 0;
	}
	fclose(fp);
	return 1;
}

int changePass(const char *ini_file) {
	FILE *fp;
	char user[33], *p;	

	if((fp = fopen(ini_file, "r+")) == NULL) {
		perror("fopen");
		return 0;
	}
	
	while(1) {
		bzero(user, 33);
		fflush(stdin);
		printf("\nUser Name:");
		scanf("%32s", user);
		getchar();
	
		if(user[0] == '\0') 
			printf("\nUser name cannot be empty");
		else
			break;
	}
		
	if(removeUser(fp, user) == 1) {
		printf("New Password:");
		p = getpass("");		
		storeUserPass(fp, user, p);
	}
	fclose(fp);
	return 1;
}


int LoadUserTable(const char *ini_file) {
	FILE *fp;
	char user[33];
	char userPass[35];
	int n;

	Trace("Loading user table...");
	
	if((fp = fopen(ini_file, "r")) == NULL) {
		Error("Cannot open %s", ini_file);
		return 0;
	}

	fseek(fp, 0L, SEEK_END);
	
	if(((int)(ftell(fp) - INI_USER_OFF) % USER_RL) != 0) {
		Error("Invalid INI file");
		fclose(fp);
		return 0;
	}
	
	n = (ftell(fp) - INI_USER_OFF) / USER_RL;

	userTableLast = 0;
	if(n == 0) {
		Trace("...done");
		Log("WARNING: No User Access Control");
		Trace("WARNING: No User Access Control");
		fclose(fp);
		return 1;
	}

	if((userTable = (struct user *) malloc(n*sizeof(struct user))) == NULL) {
		Error("Insufficient Memory");
		fclose(fp);
		return 0;
	}

	fseek(fp, INI_USER_OFF, SEEK_SET);

	des_setparity(key);	
	while(n) {
		fread(user, 1, 32, fp);
		fread(userPass, 1, 34, fp);
		user[32] = '\0';
		userPass[34] = '\0';
	
		ecb_crypt(key, user, 32, DES_DECRYPT);		
		strcpy(userTable[userTableLast].user, user);
		strcpy(userTable[userTableLast].userPass, userPass);
		userTableLast++;
		n--;
	}	
	fclose(fp);
	Trace("...done");
	return 0;
}
	
int authenticate(const char *userPass) {
	int i;
	
	if(userTableLast == 0) 
		return 1;

	for(i = 0; i < userTableLast; i++) {
		if(strncmp(userTable[i].userPass, 
						crypt(userPass, userTable[i].userPass), 34) == 0) {
			Trace("Request from %s", userTable[i].user);
			Log("Request from %s", userTable[i].user);					
			return 1;
		}
	}
	return 0;
}

extern int accept_last;
extern acc_addr acc_table[];

int IsAcceptable(struct sockaddr_in *client_addr) {
	int i;
	unsigned long client_ip;
		
	if(accept_last == 0)
		return 1;

	client_ip = ntohl((*client_addr).sin_addr.s_addr);

	for(i = 0; i < accept_last; i++) {
		if((client_ip & acc_table[i].mask) == acc_table[i].addr)
			return 1;
	}
	return 0;
}


void ReadAcceptTable(char *accept_file) {
	FILE *fp;
	unsigned long classA, classB;
	unsigned char *p, *q;
	unsigned char addr[STRING_SIZE];
	unsigned char mask[STRING_SIZE];
	char line[BUF_SIZE];
	struct in_addr sin;

	Trace("Loading accept table...");
	
	accept_last = 0;
	inet_aton("128.0.0.0", &sin);
	classA = ntohl(sin.s_addr);

	inet_aton("192.0.0.0", &sin);
	classB = ntohl(sin.s_addr);


	if((fp = fopen(accept_file, "r")) == NULL) {
		Error("Cannot open %s", accept_file);
		return;
	}

	while(fgets(line, BUF_SIZE, fp) != NULL) {
		p = line;

		while(*p && isspace(*p)) 
			p++;

		if(*p == '\0' || *p == '#')
			continue;

		if(accept_last >= ACCEPT_MAX) {
			Error("Accept table full");
			fclose(fp);
			return;
		}
		q = addr;

		while(*p && !isspace(*p))
			*q++ = *p++;
		*q = '\0';

		if(inet_aton(addr, &sin) == 0)
			continue;

		acc_table[accept_last].addr = ntohl(sin.s_addr);

		//skip WS and read mask
		
		while(*p && (*p == ' '||*p == '\t'))
			p++;

		q = mask;
		
		while(*p && !isspace(*p))
			*q++ = *p++;
		*q = '\0';

		if(inet_aton(mask, &sin) == 0) {
			//calculate default mask
			if(acc_table[accept_last].addr < classA) 
				inet_aton("255.0.0.0", &sin);
			else if(acc_table[accept_last].addr < classB) 
				inet_aton("255.255.0.0", &sin);
			else
				inet_aton("255.255.255.0", &sin);
		}
		acc_table[accept_last].mask = ntohl(sin.s_addr);

		accept_last++;
	} //while(fgets)
	fclose(fp);
	Trace("...done");
	if(accept_last == 0) {
		Trace("WARNING: No Client Access Control");
		Log("WARNING: No Client Access Control");
	}
}
