
#include "main.h"
#define NDEBUG

static int reload_tables;

void sigHupHand(int signo) {
	reload_tables = 1;
	Trace("Received SIGHUP");
	return;
}



void Reception(int client_sd, struct sockaddr_in *client_addr) {
	if(!IsAcceptable(client_addr)) {
		Trace("Access denied for client %s", inet_ntoa(client_addr->sin_addr));
		Log("Access denied for client %s", inet_ntoa(client_addr->sin_addr));
		HTTP_Error(client_sd, 403, "Forbidden", 
			"<B>403 Forbidden</B><P>Host Authentication Failed<P><I>" \
			"(This host is not permitted to access Proxy Server)</I><P>" \
			"Contact your server administrator");
		return;			
	}
	GetandSend(client_sd);
	return;
}


int InitialisePort(int port) {
	
	int proxy_sd;
	struct sockaddr_in proxy_addr;

	if((proxy_sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		Error("Unable to create socket");
		exit(1);
	}

	proxy_addr.sin_family = AF_INET;
	proxy_addr.sin_port = htons(port);
	proxy_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	bzero(&(proxy_addr.sin_zero), 8);

	if(bind(proxy_sd, (struct sockaddr *)&proxy_addr, 
				sizeof(struct sockaddr)) == -1) {
		Error("InitialisePort: Unable to Bind to port %d\n", port);
		exit(1);
	}
	
	if(listen(proxy_sd, BACKLOG) == -1) {
		Error("InitialisePort: listen failed\n");
		exit(1);
	}
	return proxy_sd;
}


void MainLoop(int proxy_sd) {
	struct sockaddr_in client_addr;
	int len, client_sd;
	pid_t chpid;
	time_t t;
   
   len = sizeof(struct sockaddr);

   while(1) {
    if((client_sd = accept(proxy_sd, 
		(struct sockaddr *)&client_addr, &len)) == -1) {
            if(errno == EINTR)
                continue;

            Error("Mainloop: cannot accept request, errno=%d", errno);
            exit(2);
    }
    conn_count++;    
 
	//Trace("Reload tables=%d", reload_tables);
	if(reload_tables == 1) {
		reload_tables = 0;
		ReadAcceptTable(ACCEPT_FILE);
		free(userTable);
		LoadUserTable(INI_FILE);
	}
      
#ifdef NDEBUG
	if((chpid = fork()) == 0) { /* child process */ 
#endif
    	close(proxy_sd);  /* close parent socket */
		
		time(&t);
      	Trace("%s  : Connection #%d from %s, pid = %d", ctime(&t), conn_count, 
		      inet_ntoa(client_addr.sin_addr), getpid());
      	Log("%s  : Connection #%d from %s", ctime(&t), conn_count, 
					  inet_ntoa(client_addr.sin_addr));
	 
        Reception(client_sd, &client_addr);

		shutdown(client_sd, 2);
		close(client_sd);

		Trace("Disconnection #%d, pid=%d\n", conn_count, getpid());
		exit(0);

#ifdef NDEBUG
	}
    else { /* parent process */
		close(client_sd);
    }
#endif
   } /* while(1) */
}
   
void ReleaseChild()
{
    /* union wait status; */

    while(wait3(NULL, WNOHANG, (struct rusage*) 0) >= 0 )
        ;
    //Trace("\nReleaseChild:\n"); 
}

extern char *optarg;
extern int optind, opterr, optopt;


int main(int argc, char *argv[]) {

	int proxy_sd;
	int proxy_port_no = HTTPPORT;
	int i;
	time_t t;

	//	InitVariables();
	strcpy(pool_dir, CACHE_DIR);


	while((i = getopt(argc, argv, "p:")) != -1) {
		switch(i) {
			case 'p':
				proxy_port_no = atoi(optarg);
				break;

		}
	}

	
	time(&t);
	Trace("%sLinProxy started, pid = %d", ctime(&t), getpid());
	
	Trace("Bind %d" , proxy_port_no);
	proxy_sd = InitialisePort(proxy_port_no);	
	Trace("Bind Success");
	
	Log("\nLinProxy started on port:%d, pid = %d", proxy_port_no, getpid());
	
	signal(SIGCHLD, ReleaseChild);
	signal(SIGHUP, sigHupHand);
	
	ReadAcceptTable(ACCEPT_FILE);
	LoadUserTable(INI_FILE);

	Trace("Waiting for Connections ..............\n");
	MainLoop(proxy_sd);
}
	
